import math
import sys
from numbers import Rational
from typing import (Callable,
                    Sequence,
                    Tuple,
                    TypeVar)

from cfractions import Fraction

from .hints import RawConstant

_T1 = TypeVar('_T1')
_T2 = TypeVar('_T2')

BASE = Fraction(10)


def ceil_half(value: int) -> int:
    return -(-value // 2)


def digits_count(value: int) -> int:
    return len('%i' % abs(value))


def identity(value: _T1) -> _T1:
    return value


if sys.version_info < (3, 9):
    def lcm(left: int, right: int) -> int:
        left, right = abs(left), abs(right)
        return min(left, right) * (max(left, right) // math.gcd(left, right))
else:
    lcm = math.lcm


def positiveness_to_sign(flag: bool) -> int:
    return 2 * flag - 1


def rational_sqrt_lower_bound(value: RawConstant) -> Fraction:
    assert isinstance(value, Rational), value
    return Fraction(sqrt_floor(value.numerator * value.denominator),
                    value.denominator)


def rational_sqrt_upper_bound(value: RawConstant) -> Fraction:
    assert isinstance(value, Rational), value
    return Fraction(sqrt_ceil(value.numerator * value.denominator),
                    value.denominator)


def sqrt_ceil(value: int) -> int:
    value_sqrt_floor = sqrt_floor(value)
    return value_sqrt_floor + (value != value_sqrt_floor * value_sqrt_floor)


if sys.version_info < (3, 8):
    def sqrt_floor(value: int) -> int:
        if value > 0:
            candidate = 1 << ((value.bit_length() + 1) >> 1)
            while True:
                next_candidate = (candidate + value // candidate) >> 1
                if next_candidate >= candidate:
                    return candidate
                candidate = next_candidate
        else:
            assert not value, 'Argument should be non-negative.'
            return 0
else:
    sqrt_floor = math.isqrt


def perfect_sqrt(value: int, alternative: int = 1) -> int:
    candidate = sqrt_floor(value)
    return candidate if candidate * candidate == value else alternative


to_square_free: Callable[[int], int]

try:
    from symba._symba import to_square_free
except ImportError:
    def to_square_free(value: int) -> int:
        while value % 4 == 0:
            value //= 4
        factor_candidate_squared = 1
        for factor_candidate_base in range(1, value, 2):
            factor_candidate_squared += 4 * factor_candidate_base + 4
            if factor_candidate_squared > value:
                break
            quotient, remainder = divmod(value, factor_candidate_squared)
            while not remainder:
                value = quotient
                quotient, remainder = divmod(value, factor_candidate_squared)
        return value


def transpose(
        pairs_sequence: Sequence[Tuple[_T1, _T2]]
) -> Tuple[Sequence[_T1], Sequence[_T2]]:
    first_coordinates, second_coordinates = zip(*pairs_sequence)
    return first_coordinates, second_coordinates



from __future__ import annotations

import math
from collections import defaultdict
from functools import reduce
from numbers import (Rational,
                     Real)
from typing import (Any,
                    Callable,
                    DefaultDict,
                    Iterable,
                    List,
                    Optional,
                    Sequence,
                    Tuple,
                    Union,
                    cast,
                    overload)

from reprit.base import generate_repr

from .constant import (ONE,
                       RAW_ZERO,
                       ZERO,
                       Constant,
                       Finite,
                       FiniteNonZero,
                       Infinite,
                       Zero,
                       to_constant)
from .expression import Expression
from .hints import (RawConstant,
                    RawFinite,
                    RawUnbound)
from .term import Term
from .utils import (BASE,
                    digits_count,
                    lcm,
                    positiveness_to_sign,
                    to_square_free,
                    transpose)


class Form(Expression):
    """Represents sum of square roots."""

    @classmethod
    def from_components(cls,
                        terms: List[Term],
                        tail: Union[FiniteNonZero, Zero] = ZERO) -> Expression:
        queue = sorted(terms,
                       key=abs,
                       reverse=True)
        arguments_scales: DefaultDict[
            Union[FiniteNonZero, Form, Term], Finite] = defaultdict(Zero)
        while queue:
            min_term = queue.pop()
            min_term_argument = min_term.argument
            arguments_scales[min_term_argument] += min_term.scale
            next_queue = []
            for term in queue:
                arguments_ratio = term.argument / min_term_argument
                arguments_ratio_sqrt = arguments_ratio.perfect_sqrt()
                if arguments_ratio_sqrt.square() == arguments_ratio:
                    arguments_scales[min_term_argument] += (
                            term.scale * arguments_ratio_sqrt
                    )
                else:
                    next_queue.append(term)
            queue = next_queue
        terms = [Term(scale, argument)
                 for argument, scale in arguments_scales.items()
                 if not isinstance(scale, Zero)]
        return ((cls(terms, tail) if tail or len(terms) > 1 else terms[0])
                if terms
                else tail)

    __slots__ = 'tail', 'terms'

    def __init__(self, terms: List[Term], tail: Finite = ZERO) -> None:
        self.tail, self.terms = tail, terms

    @property
    def degree(self) -> int:
        return max(term.degree for term in self.terms)

    def extract_common_denominator(self) -> Tuple[int, Form]:
        terms_common_denominators, _ = transpose(
                [term.extract_common_denominator() for term in self.terms]
        )
        tail_denominator, _ = self.tail.extract_common_denominator()
        common_denominator = reduce(lcm, terms_common_denominators,
                                    tail_denominator)
        return (common_denominator,
                self._scale(FiniteNonZero(common_denominator)))

    def extract_common_numerator(self) -> Tuple[int, Form]:
        terms_common_numerators, _ = transpose([term.extract_common_numerator()
                                                for term in self.terms])
        tail_numerator, _ = self.tail.extract_common_numerator()
        common_numerator = reduce(math.gcd, terms_common_numerators,
                                  tail_numerator)
        return common_numerator, self / common_numerator

    def inverse(self) -> Union[Finite, Form, Term]:
        common_denominator, integer_form = self.extract_common_denominator()
        numerator, denominator = (
            Factorization(tail=to_constant(common_denominator)),
            Factorization.from_form(integer_form)
        )
        while denominator.factors:
            max_factor = max(denominator.factors)
            max_factorization = denominator.factors.pop(max_factor)
            numerator = numerator.multiply(
                    denominator
                    - max_factorization.multiply_by_factor(max_factor)
            )
            denominator = (denominator.square()
                           - max_factorization.square() * max_factor.square())
        return numerator.scale_non_zero(denominator.tail.inverse()).express()

    def is_positive(self) -> bool:
        components = (*self.terms, self.tail) if self.tail else self.terms
        positive: List[Expression] = []
        negative: List[Expression] = []
        for component in components:
            (positive
             if component.is_positive()
             else negative).append(component)
        if not (positive and negative):
            return not negative
        positive_squares_sum, negative_squares_sum = (
            sum(component.square() for component in positive),
            sum(component.square() for component in negative)
        )
        assert isinstance(positive_squares_sum, Expression), (
            positive_squares_sum
        )
        assert isinstance(negative_squares_sum, Expression), (
            negative_squares_sum
        )
        return ((len(positive) * positive_squares_sum
                 - negative_squares_sum).is_positive()
                and ((positive_squares_sum
                      - len(negative) * negative_squares_sum).is_positive()
                     or self.lower_bound() >= RAW_ZERO))

    def lower_bound(self) -> RawConstant:
        common_denominator, form = self.extract_common_denominator()
        scale = BASE ** form.significant_digits_count()
        return (sum([(scale * term).lower_bound() for term in form.terms],
                    (scale * form.tail).lower_bound())
                / (common_denominator * scale))

    def perfect_sqrt(self) -> Expression:
        if self.degree != 1:
            raise ValueError('Unsupported value: {!r}.'.format(self))
        denominator, integer_form = self.extract_common_denominator()
        if not self.tail:
            arguments_gcd = form_arguments_gcd(integer_form)
            if arguments_gcd != 1:
                common_term = Term(ONE, FiniteNonZero(arguments_gcd))
                normalized_form = integer_form / common_term
                normalized_sqrt = normalized_form.perfect_sqrt()
                if normalized_sqrt.square() == normalized_form:
                    result = (Term(ONE / denominator, common_term)
                              * normalized_sqrt)
                    assert isinstance(result, Expression), result
                    return result
        elif len(self.terms) == 1:
            term, = integer_form.terms
            discriminant = integer_form.tail.square() - term.square()
            if discriminant.is_positive():
                # checking if the form can be represented as
                # ``(a * sqrt(x) + b * sqrt(y)) ** 2``,
                # where
                # ``a, b, x, y`` are rational,
                # ``x, y`` are non-equal
                discriminant_sqrt = discriminant.perfect_sqrt()
                if discriminant_sqrt.square() == discriminant:
                    scale = ONE / denominator
                    return (positiveness_to_sign(term.is_positive())
                            * Term.from_components(scale,
                                                   (integer_form.tail
                                                    - discriminant_sqrt) / 2)
                            + Term.from_components(scale,
                                                   (integer_form.tail
                                                    + discriminant_sqrt) / 2))
            else:
                # checking if the form can be represented as
                # ``((sqrt(a) + b * sqrt(c)) * sqrt(sqrt(x))) ** 2``,
                # where
                # ``a, b, c, x`` are non-zero rational,
                # ``a, c, x`` are positive,
                # ``a, c`` are non-equal,
                # ``a * c`` is a perfect square
                discriminant /= -term.argument
                discriminant_sqrt = discriminant.perfect_sqrt()
                if discriminant_sqrt.square() == discriminant:
                    scale = ONE / denominator
                    sub_term = Term(ONE, term.argument)
                    return (positiveness_to_sign(self.tail.is_positive())
                            * Term(scale,
                                   (term.scale - discriminant_sqrt) / 2
                                   * sub_term)
                            + Term(scale,
                                   (term.scale + discriminant_sqrt) / 2
                                   * sub_term))
        else:
            lesser_part, greater_part = _split_form(integer_form)
            discriminant = greater_part.square() - lesser_part.square()
            if discriminant.is_positive():
                discriminant_is_constant = not discriminant.degree
                discriminant_sqrt = (Term.from_components(ONE, discriminant)
                                     if discriminant_is_constant
                                     else discriminant.perfect_sqrt())
                if (discriminant_is_constant
                        or discriminant_sqrt.square() == discriminant):
                    addend = greater_part + discriminant_sqrt
                    if (not isinstance(addend, Form)
                            or (len(addend.terms) + bool(addend.tail)
                                < len(self.terms) + bool(self.tail))):
                        result = ((lesser_part + addend)
                                  / Term.from_components(
                                        FiniteNonZero(denominator), 2 * addend
                                ))
                        assert isinstance(result, Expression), result
                        return result
        common_denominator, integer_form = self.extract_common_denominator()
        common_numerator, _ = integer_form.extract_common_numerator()
        return (
                FiniteNonZero(common_numerator) / common_denominator
        ).perfect_sqrt()

    def significant_digits_count(self) -> int:
        return (max(max(term.significant_digits_count()
                        for term in self.terms),
                    self.tail.significant_digits_count())
                + digits_count(len(self.terms) + bool(self.tail)) + 1)

    def square(self) -> Expression:
        terms = ([]
                 if isinstance(self.tail, Zero)
                 else [(2 * self.tail) * term for term in self.terms])
        tail = (self.tail.square()
                + _sift_components([2 * (self.terms[step] * self.terms[index])
                                    for step in range(1, len(self.terms))
                                    for index in range(step)]
                                   + [term.square() for term in self.terms],
                                   terms))
        return Form.from_components(terms, tail)

    def upper_bound(self) -> RawConstant:
        common_denominator, form = self.extract_common_denominator()
        scale = BASE ** form.significant_digits_count()
        return (sum([(scale * term).upper_bound() for term in form.terms],
                    (scale * form.tail).upper_bound())
                / (common_denominator * scale))

    @overload
    def __add__(self, other: RawFinite) -> Union[Form, Term]:
        ...

    @overload
    def __add__(self, other: RawUnbound) -> Union[Form, Infinite, Term]:
        ...

    @overload
    def __add__(self, other: FiniteNonZero) -> Union[Form, Term]:
        ...

    @overload
    def __add__(self, other: Infinite) -> Infinite:
        ...

    @overload
    def __add__(self, other: Zero) -> Form:
        ...

    @overload
    def __add__(self, other: Form) -> Union[Finite, Form, Term]:
        ...

    @overload
    def __add__(self, other: Any) -> Any:
        ...

    def __add__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._add_constant(other)
                if isinstance(other, (FiniteNonZero, Infinite, Zero))
                else (self._add_term(other)
                      if isinstance(other, Term)
                      else (Form.from_components(self.terms + other.terms,
                                                 self.tail + other.tail)
                            if isinstance(other, Form)
                            else NotImplemented)))

    @overload
    def __eq__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __eq__(self, other: Any) -> Any:
        ...

    def __eq__(self, other: Any) -> Any:
        return (self is other
                or (self.tail == other.tail
                    and len(self.terms) == len(other.terms)
                    and set(self.terms) == set(other.terms))
                if isinstance(other, Form)
                else (False
                      if isinstance(other, (Expression, Rational, float))
                      else NotImplemented))

    def __hash__(self) -> int:
        return hash((frozenset(self.terms), self.tail))

    @overload
    def __mul__(self, other: RawFinite) -> Union[Form, Zero]:
        ...

    @overload
    def __mul__(self, other: RawUnbound) -> Union[Form, Infinite, Zero]:
        ...

    @overload
    def __mul__(self, other: Union[FiniteNonZero, Term]) -> Form:
        ...

    @overload
    def __mul__(self, other: Infinite) -> Infinite:
        ...

    @overload
    def __mul__(self, other: Zero) -> Zero:
        ...

    @overload
    def __mul__(self, other: Form) -> Union[FiniteNonZero, Form]:
        ...

    @overload
    def __mul__(self, other: Expression) -> Expression:
        ...

    @overload
    def __mul__(self, other: Any) -> Any:
        ...

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._multiply_by_constant(other)
                if isinstance(other, Constant)
                else (self._multiply_by_term(other)
                      if isinstance(other, Term)
                      else (self._multiply_by_form(other)
                            if isinstance(other, Form)
                            else NotImplemented)))

    def __neg__(self) -> Form:
        return Form([-term for term in self.terms],
                    tail=-self.tail)

    @overload
    def __radd__(self, other: Union[RawConstant, Expression]) -> Expression:
        ...

    @overload
    def __radd__(self, other: Any) -> Any:
        ...

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._add_constant(other)
                if isinstance(other, (FiniteNonZero, Infinite, Zero))
                else (self._add_term(other)
                      if isinstance(other, Term)
                      else NotImplemented))

    __repr__ = generate_repr(__init__)

    @overload
    def __rmul__(self, other: RawFinite) -> Union[Form, Zero]:
        ...

    @overload
    def __rmul__(self, other: RawUnbound) -> Union[Form, Infinite, Zero]:
        ...

    @overload
    def __rmul__(self, other: FiniteNonZero) -> Form:
        ...

    @overload
    def __rmul__(self, other: Term) -> Form:
        ...

    @overload
    def __rmul__(self, other: Infinite) -> Infinite:
        ...

    @overload
    def __rmul__(self, other: Zero) -> Zero:
        ...

    @overload
    def __rmul__(self, other: Expression) -> Expression:
        ...

    @overload
    def __rmul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._multiply_by_constant(other)
                if isinstance(other, Constant)
                else (self._multiply_by_term(other)
                      if isinstance(other, Term)
                      else NotImplemented))

    def __str__(self) -> str:
        return (str(self.terms[0])
                + (' ' + ' '.join(map(_to_signed_value, self.terms[1:]))
                   if len(self.terms) > 1
                   else '')
                + (''
                   if isinstance(self.tail, Zero)
                   else ' ' + _to_signed_value(self.tail)))

    def _add_constant(self, other: Union[Finite, Infinite]) -> Expression:
        tail = self.tail + other
        return (tail
                if isinstance(tail, Infinite)
                else (Form(self.terms, tail)
                      if tail or len(self.terms) > 1
                      else self.terms[0]))

    def _add_term(self, other: Term) -> Expression:
        return Form.from_components(self.terms + [other], self.tail)

    def _multiply_by_constant(self, other: Constant) -> Expression:
        return (self._scale(other)
                if isinstance(other, FiniteNonZero)
                else ((other if self.is_positive() else -other)
                      if isinstance(other, Infinite)
                      else other))

    def _multiply_by_form(self, other: Form) -> Expression:
        if self == other:
            return self.square()
        tail, other_tail = self.tail, other.tail
        terms = (([]
                  if isinstance(other_tail, Zero)
                  else [term * other_tail for term in self.terms])
                 + ([]
                    if isinstance(tail, Zero)
                    else [other_term * tail for other_term in other.terms]))
        tail = (tail * other_tail
                + _sift_components([term * other_term
                                    for term in self.terms
                                    for other_term in other.terms],
                                   terms))
        return Form.from_components(terms, tail)

    def _multiply_by_term(self, other: Term) -> Expression:
        terms = [] if isinstance(self.tail, Zero) else [other * self.tail]
        return Form(terms,
                    _sift_components([term * other for term in self.terms],
                                     terms))

    def _scale(self, other: FiniteNonZero) -> Form:
        return (self
                if other == ONE
                else Form([term * other for term in self.terms],
                          self.tail * other))


def form_arguments_gcd(integer_form: Form) -> int:
    result, _, coprime_indices = _split_integers_by_gcd(
            [_evaluate_integer_term_argument(term)
             for term in sorted(integer_form.terms,
                                key=_term_key)]
    )
    return 1 if coprime_indices else result


def _split_form(integer_form: Form) -> Tuple[Form, Form]:
    terms: Sequence[Term] = integer_form.terms
    if len(terms) == 3:
        first_term, *rest_terms = terms
        return Form(rest_terms), Form([first_term], integer_form.tail)
    surds, terms = zip(*sorted([(_integer_term_to_surd(term), term)
                                for term in terms]))
    cocomposite_indices, coprime_indices = _split_integers(surds)
    return ((Form([terms[index] for index in cocomposite_indices]),
             Form([terms[index] for index in coprime_indices],
                  integer_form.tail)))


def _integer_term_to_surd(term: Term) -> int:
    return to_square_free(_evaluate_integer_term_argument(term))


def _evaluate_integer_term_argument(term: Term) -> int:
    argument = term.argument
    assert isinstance(argument, FiniteNonZero), argument
    return argument.raw.numerator


def _split_integers(values: Sequence[int]) -> Tuple[List[int], List[int]]:
    gcd, cocomposite_indices, coprime_indices = _split_integers_by_gcd(values)
    if not coprime_indices:
        _, cocomposite_indices, coprime_indices = _split_integers_by_gcd(
                value // gcd for value in values
        )
    return cocomposite_indices, coprime_indices


def _split_integers_by_gcd(
        integers: Iterable[int]
) -> Tuple[int, List[int], List[int]]:
    iterator = iter(integers)
    gcd = next(iterator)
    cocomposite_indices, coprime_indices, gcd, start = (
        ([1], [0], next(iterator), 2) if gcd == 1 else ([0], [], gcd, 1)
    )
    for index, value in enumerate(iterator,
                                  start=start):
        value_gcd = math.gcd(gcd, value)
        if value_gcd == 1:
            coprime_indices.append(index)
        else:
            gcd = value_gcd
            cocomposite_indices.append(index)
    return gcd, cocomposite_indices, coprime_indices


class Factor:
    def express(self) -> Term:
        return self._term

    def square(self) -> Expression:
        return self._term.argument

    argument: Union[FiniteNonZero, Form, Term]
    degree: int
    _term: Term

    __slots__ = 'argument', 'degree', '_term'

    def __init__(self,
                 argument: Union[FiniteNonZero, Form, Term],
                 degree: int) -> None:
        self.argument, self.degree = argument, degree
        term = argument
        for _ in range(degree):
            term = Term(ONE, term)
        assert isinstance(term, Term)
        self._term = term

    @overload
    def __eq__(self, other: 'Factor') -> bool:
        ...

    @overload
    def __eq__(self, other: Any) -> Any:
        ...

    def __eq__(self, other: 'Factor') -> bool:
        return (self.degree == other.degree and self.argument == other.argument
                if isinstance(other, Factor)
                else NotImplemented)

    def __hash__(self) -> int:
        return hash((self.argument, self.degree))

    def __lt__(self, other: 'Factor') -> bool:
        return ((self.express().degree, self.express().argument)
                < (other.express().degree, other.express().argument))

    __repr__ = generate_repr(__init__)

    def __str__(self) -> str:
        return str(self.express())


class Factorization:
    @classmethod
    def from_factor(cls, factor: Factor) -> Factorization:
        result = cls()
        result.factors[factor].tail = ONE
        return result

    @classmethod
    def from_form(cls, form: Form) -> Factorization:
        result = cls(tail=form.tail)
        factors = result.factors
        for term in form.terms:
            _populate_factors(factors, term)
        return result

    @classmethod
    def from_term(cls, term: Term) -> Factorization:
        result = cls()
        _populate_factors(result.factors, term)
        return result

    __slots__ = 'factors', 'tail'

    def __init__(self,
                 factors: Optional[DefaultDict[Factor, Factorization]]
                 = None,
                 tail: Union[FiniteNonZero, Zero] = ZERO) -> None:
        self.factors, self.tail = (defaultdict(Factorization)
                                   if factors is None
                                   else factors,
                                   tail)

    def express(self) -> Union[Finite, Form, Term]:
        return sum([factor.express() * factorization.express()
                    for factor, factorization in self.factors.items()],
                   self.tail)

    def multiply(self: Factorization, other: Factorization) -> Factorization:
        factors, tail = self.factors, self.tail
        other_factors, other_tail = other.factors, other.tail
        result = self.scale(other.tail) + other.scale(self.tail)
        result.tail /= 2
        for factor, factorization in factors.items():
            for other_factor, other_factorization in other_factors.items():
                result_factorization = (
                    factorization.multiply(other_factorization)
                )
                if factor == other_factor:
                    squared_factor = factor.square()
                    assert isinstance(squared_factor,
                                      (FiniteNonZero, Form, Term)), (
                        squared_factor
                    )
                    result += (
                        result_factorization.scale_non_zero(squared_factor)
                        if isinstance(squared_factor, FiniteNonZero)
                        else (result_factorization
                              .multiply_by_form(squared_factor)
                              if isinstance(squared_factor, Form)
                              else (result_factorization
                                    .multiply_by_term(squared_factor)))
                    )
                else:
                    max_factor, min_factor = ((other_factor, factor)
                                              if factor < other_factor
                                              else (factor, other_factor))
                    result.factors[max_factor] += (
                        result_factorization.multiply_by_factor(min_factor))
        return result

    def multiply_by_factor(self, factor: Factor) -> Factorization:
        return self.multiply(Factorization.from_factor(factor))

    def multiply_by_form(self, form: Form) -> Factorization:
        return sum([self.multiply_by_term(term) for term in form.terms],
                   self.scale(form.tail))

    def multiply_by_term(self, term: Term) -> Factorization:
        return self.multiply(Factorization.from_term(term))

    def scale(self, scale: Union[FiniteNonZero, Zero]) -> Factorization:
        return (Factorization()
                if isinstance(scale, Zero)
                else self.scale_non_zero(scale))

    def scale_non_zero(self, scale: FiniteNonZero) -> Factorization:
        result = Factorization(tail=self.tail * scale)
        factors = result.factors
        for factor, factorization in self.factors.items():
            factors[factor] = factorization.scale_non_zero(scale)
        return result

    def square(self, _two: FiniteNonZero = FiniteNonZero(2)) -> Factorization:
        result = self.scale(_two * self.tail)
        result.tail /= 2
        factorizations = tuple(self.factors.items())
        for offset, (factor, factorization) in enumerate(factorizations,
                                                         start=1):
            result += factorization.square() * factor.square()
            for next_index in range(offset, len(factorizations)):
                next_factor, next_factorization = factorizations[next_index]
                max_factor, min_factor = ((next_factor, factor)
                                          if factor < next_factor
                                          else (factor, next_factor))
                result.factors[max_factor] += (factorization
                                               .scale_non_zero(_two)
                                               .multiply_by_factor(min_factor)
                                               .multiply(next_factorization))
        return result

    def __add__(self, other: Factorization) -> Factorization:
        if not (self and other):
            return self or other
        factors, rest_factors = ((other.factors.copy(), self.factors)
                                 if len(self.factors) < len(other.factors)
                                 else (self.factors.copy(), other.factors))
        for factor, factorization in rest_factors.items():
            factors[factor] += factorization
        return Factorization(factors, self.tail + other.tail)

    def __bool__(self) -> bool:
        return bool(self.tail or any(self.factors.values()))

    def __iadd__(self, other: Factorization) -> Factorization:
        if not self:
            self.factors, self.tail = other.factors.copy(), other.tail
        elif other:
            factors, rest_factors = ((other.factors.copy(), self.factors)
                                     if len(self.factors) < len(other.factors)
                                     else (self.factors, other.factors))
            for factor, factorization in rest_factors.items():
                factors[factor] += factorization
            self.factors = factors
            self.tail += other.tail
        return self

    def __len__(self) -> int:
        return bool(self.tail) + sum(map(len, self.factors.values()))

    @overload
    def __mul__(self, other: Expression) -> Factorization:
        ...

    @overload
    def __mul__(self, other: Any) -> Any:
        ...

    def __mul__(self, other: Any) -> Any:
        return (self.scale(other)
                if isinstance(other, FiniteNonZero)
                else (self.multiply_by_term(other)
                      if isinstance(other, Term)
                      else (self.multiply_by_form(other)
                            if isinstance(other, Form)
                            else NotImplemented)))

    def __neg__(self) -> Factorization:
        result = Factorization(tail=-self.tail)
        factors = result.factors
        for factor, factorization in self.factors.items():
            factors[factor] = -factorization
        return result

    __repr__ = generate_repr(__init__)

    __rmul__ = __mul__

    def __str__(self) -> str:
        head = ' + '.join([
            '{} * {}'.format(
                    factor,
                    cast(
                            Callable[[Factorization], str],
                            (str
                             if (len(factorization) == 1
                                 and (factorization.tail >= 0
                                      and
                                      all(sub_factorization.tail >= 0
                                          for sub_factorization
                                          in factorization.factors.values())))
                             else '({})'.format)
                    )(factorization)
            )
            for factor, factorization in self.factors.items()
            if factorization
        ])
        return ((head + (' + ' + str(self.tail) if self.tail else ''))
                if head
                else str(self.tail))

    def __sub__(self, other: Factorization) -> Factorization:
        return self + (-other)


def _factor_term(term: Term) -> Iterable[Factor]:
    queue: List[Tuple[int, Term]] = [(0, term)]
    while queue:
        degree, step = queue.pop()
        if degree and step.scale != 1:
            yield Factor(step.scale, degree)
        argument = step.argument
        if isinstance(argument, Term):
            queue.append((degree + 1, argument))
        elif isinstance(argument, Form):
            common_denominator, argument = (argument
                                            .extract_common_denominator())
            common_numerator, argument = argument.extract_common_numerator()
            if not (common_numerator == 1 == common_denominator):
                yield Factor(
                        FiniteNonZero(common_numerator) / common_denominator,
                        degree + 1
                )
            yield Factor(argument, degree + 1)
        else:
            yield Factor(argument, degree + 1)


def _populate_factors(children: DefaultDict[Factor, Factorization],
                      term: Term) -> None:
    *rest_factors, first_factor = _factor_term(term)
    last_factorization = children[first_factor]
    for factor in rest_factors:
        last_factorization = last_factorization.factors[factor]
    last_factorization.tail += term.scale


def _sift_components(components: Iterable[Expression],
                     terms: List[Term]) -> Finite:
    tail: Finite = ZERO
    for component in components:
        if isinstance(component, Term):
            terms.append(component)
        elif isinstance(component, Form):
            tail += component.tail
            terms.extend(component.terms)
        else:
            assert isinstance(component, FiniteNonZero), component
            tail += component
    return tail


def _term_key(term: Term) -> Tuple[int, Expression]:
    return term.degree, term.argument


def _to_signed_value(value: Union[FiniteNonZero, Term]) -> str:
    return '+ ' + str(value) if value.is_positive() else '- ' + str(-value)



from __future__ import annotations

import math
from abc import (ABC,
                 abstractmethod)
from numbers import Rational
from typing import (Any,
                    Optional,
                    Tuple,
                    TypeVar,
                    Union,
                    cast,
                    overload)

from cfractions import Fraction

from .hints import (RawConstant,
                    RawFinite,
                    RawUnbound)
from .utils import BASE

_Self = TypeVar('_Self',
                bound='Expression')


class Expression(ABC):
    __slots__ = ()

    @property
    @abstractmethod
    def degree(self) -> int:
        """Returns degree of the expression."""

    @abstractmethod
    def extract_common_denominator(self) -> Tuple[int, Expression]:
        """
        Returns a pair of the common denominator of the expression
        and the rest of the expression.
        """

    @abstractmethod
    def extract_common_numerator(self) -> Tuple[int, Expression]:
        """
        Returns a pair of the common numerator of the expression
        and the rest of the expression.
        """

    @abstractmethod
    def inverse(self) -> Expression:
        """Returns the expression inverted."""

    @abstractmethod
    def is_positive(self) -> bool:
        """Checks if the expression is positive."""

    @abstractmethod
    def lower_bound(self) -> RawConstant:
        """Returns lower bound of the expression."""

    @abstractmethod
    def perfect_sqrt(self) -> Expression:
        """Returns perfect square root part of the expression."""

    @abstractmethod
    def significant_digits_count(self) -> int:
        """Returns significant digits count of the expression."""

    @abstractmethod
    def square(self) -> Expression:
        """Returns the expression squared."""

    @abstractmethod
    def upper_bound(self) -> RawConstant:
        """Returns upper bound of the expression."""

    def __abs__(self) -> Expression:
        """Returns an absolute value of the expression."""
        return self if self.is_positive() else -self

    @abstractmethod
    def __add__(self, other: Union[RawConstant, Expression]) -> Expression:
        """Returns sum of the expression with the other."""

    def __ceil__(self) -> int:
        """Return the ceiling of the expression."""
        return math.ceil(self.upper_bound())

    def __floor__(self) -> int:
        """Return the floor of the expression."""
        return math.floor(self.lower_bound())

    def __floordiv__(self, other: Union[RawConstant, Expression]) -> int:
        """Returns quotient of the division of the expression by the other."""
        return (self / other).__floor__()

    @overload
    def __ge__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __ge__(self, other: Any) -> Any:
        ...

    def __ge__(self, other: Any) -> Any:
        """Checks if the expression is greater than or equal to the other."""
        from .constant import (Infinite,
                               to_constant)
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other <= self
                 if isinstance(other, Infinite)
                 else not (other - self).is_positive())
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __gt__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __gt__(self, other: Any) -> Any:
        ...

    def __gt__(self, other: Any) -> Any:
        """Checks if the expression is greater than the other."""
        from .constant import (Infinite,
                               to_constant)
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other < self
                 if isinstance(other, Infinite)
                 else (self - other).is_positive())
                if isinstance(other, Expression)
                else NotImplemented)

    @abstractmethod
    def __hash__(self) -> int:
        """Returns hash value of the expression."""

    @overload
    def __le__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __le__(self, other: Any) -> Any:
        ...

    def __le__(self, other: Any) -> Any:
        """Checks if the expression is lower than or equal to the other."""
        from .constant import (Infinite,
                               to_constant)
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other >= self
                 if isinstance(other, Infinite)
                 else not (self - other).is_positive())
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __lt__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __lt__(self, other: Any) -> Any:
        ...

    def __lt__(self, other: Any) -> Any:
        """Checks if the expression is lower than the other."""
        from .constant import (Infinite,
                               to_constant)
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other > self
                 if isinstance(other, Infinite)
                 else (other - self).is_positive())
                if isinstance(other, Expression)
                else NotImplemented)

    def __mod__(self, other: Union[RawConstant, Expression]) -> Expression:
        """Returns remainder of the division of the expression by the other."""
        return self - other * (self // other)

    @abstractmethod
    def __mul__(self, other: Union[RawConstant, Expression]) -> Expression:
        """Returns multiplication of the expression with the other."""

    @abstractmethod
    def __neg__(self) -> Expression:
        """Returns the expression negated."""

    def __pos__(self) -> Expression:
        """Returns the expression positive."""
        return self

    def __pow__(self, exponent: int) -> Expression:
        """Returns the expression raised to the given exponent."""
        if not isinstance(exponent, int):
            return NotImplemented
        from .constant import ONE
        if not exponent:
            return ONE
        result, step = ONE, self
        if exponent < 0:
            exponent, step = -exponent, step.inverse()
        while exponent > 1:
            if exponent & 1:
                result *= step
            step = step.square()
            exponent >>= 1
        result *= step
        return result

    @abstractmethod
    def __radd__(self, other: Union[RawConstant, Expression]) -> Expression:
        """Returns sum of the other with the expression."""

    def __rfloordiv__(self, other: Union[RawConstant, Expression]) -> int:
        """Returns quotient of the division of the other by the expression."""
        return (other / self).__floor__()

    def __rmod__(self, other: Union[RawConstant, Expression]) -> Expression:
        """Returns remainder of the division of the other by the expression."""
        return other - self * (other // self)

    @abstractmethod
    def __rmul__(self, other: Union[RawConstant, Expression]) -> Expression:
        """Returns multiplication of the other with the expression."""

    def __round__(self,
                  precision: Optional[int] = None) -> Union[int, Fraction]:
        """Returns the expression rounded to the given precision."""
        scale = BASE ** (1 if precision is None else precision + 1)
        return round(cast(Fraction, int(scale * self) / scale), precision)

    @overload
    def __rsub__(self, other: Union[RawConstant, Expression]) -> Expression:
        ...

    @overload
    def __rsub__(self, other: Any) -> Any:
        ...

    def __rsub__(self, other: Any) -> Any:
        """Returns difference of the other with the expression."""
        from .constant import to_constant
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (other + (-self)
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __rtruediv__(self,
                     other: Union[Expression, RawConstant]) -> Expression:
        ...

    @overload
    def __rtruediv__(self, other: Any) -> Any:
        ...

    def __rtruediv__(self, other: Any) -> Any:
        """Returns division of the other by the expression."""
        return (self.inverse() * other
                if isinstance(other, (Rational, float))
                else NotImplemented)

    @overload
    def __sub__(self, other: Union[RawConstant, Expression]) -> Expression:
        ...

    @overload
    def __sub__(self, other: Any) -> Any:
        ...

    def __sub__(self, other: Any) -> Any:
        """Returns difference of the expression with the other."""
        from .constant import to_constant
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self + (-other)
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __truediv__(self: _Self, other: RawFinite) -> _Self:
        ...

    @overload
    def __truediv__(self, other: Union[Expression, RawUnbound]) -> Expression:
        ...

    def __truediv__(self, other: Any) -> Any:
        """Returns division of the expression by the other."""
        from .constant import to_constant
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self * other.inverse()
                if isinstance(other, Expression)
                else NotImplemented)

    def __trunc__(self) -> int:
        """Returns the expression truncated to a nearest-to-zero integer."""
        return self.__floor__() if self.is_positive() else self.__ceil__()



from __future__ import annotations

import math
from abc import abstractmethod
from numbers import Rational
from typing import (Any,
                    NoReturn,
                    Tuple,
                    TypeVar,
                    Union,
                    overload)

from cfractions import Fraction
from reprit.base import generate_repr

from .expression import Expression
from .hints import (RawConstant,
                    RawFinite,
                    RawUnbound)
from .utils import (digits_count,
                    identity,
                    perfect_sqrt,
                    positiveness_to_sign)

RAW_ZERO = Fraction(0)
RAW_ONE = Fraction(1)


class Constant(Expression):
    @property
    def degree(self) -> int:
        return 0

    @property
    @abstractmethod
    def raw(self) -> RawConstant:
        """Returns value of the constant."""

    def lower_bound(self) -> RawConstant:
        return self.raw

    upper_bound = lower_bound

    def __eq__(self, other: Any) -> Any:
        return (self.raw == other
                if isinstance(other, (Rational, float))
                else (isinstance(other, Constant)
                      and self.raw == other.raw
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __hash__(self) -> int:
        return hash(self.raw)

    def __str__(self) -> str:
        return str(self.raw)


_Expression = TypeVar('_Expression',
                      bound=Expression)


class Zero(Constant):
    """Represents zero."""

    @property
    def raw(self) -> Fraction:
        return RAW_ZERO

    def extract_common_denominator(self) -> Tuple[int, Zero]:
        return 1, self

    def extract_common_numerator(self) -> Tuple[int, FiniteNonZero]:
        return 0, ONE

    def inverse(self) -> NoReturn:
        raise ZeroDivisionError()

    def is_positive(self) -> bool:
        return False

    def perfect_sqrt(self) -> Zero:
        return self

    def significant_digits_count(self) -> int:
        return 1

    def square(self) -> Zero:
        return self

    def __new__(cls) -> Zero:
        return super().__new__(cls)

    @overload
    def __add__(self, other: RawFinite) -> Finite:
        ...

    @overload
    def __add__(self, other: RawUnbound) -> Union[Finite, Infinite]:
        ...

    @overload
    def __add__(self, other: Zero) -> Zero:
        ...

    @overload
    def __add__(self, other: _Expression) -> _Expression:
        ...

    @overload
    def __add__(self, other: Any) -> Any:
        ...

    def __add__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (other
                if isinstance(other, Expression)
                else NotImplemented)

    def __bool__(self) -> bool:
        return False

    @overload
    def __mul__(self, other: Union[RawConstant, Zero]) -> Zero:
        ...

    @overload
    def __mul__(self, other: Any) -> Any:
        ...

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self
                if isinstance(other, Zero)
                else (other * self
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __neg__(self) -> Zero:
        return self

    @overload
    def __radd__(self, other: RawFinite) -> Finite:
        ...

    @overload
    def __radd__(self, other: RawUnbound) -> Union[Finite, Infinite]:
        ...

    @overload
    def __radd__(self, other: _Expression) -> _Expression:
        ...

    @overload
    def __radd__(self, other: Any) -> Any:
        ...

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (other
                if isinstance(other, Expression)
                else NotImplemented)

    __repr__ = generate_repr(__new__)

    @overload
    def __rmul__(self, other: Union[RawConstant, Zero]) -> Zero:
        ...

    @overload
    def __rmul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        return (to_constant(other) * self
                if isinstance(other, (Rational, float))
                else NotImplemented)


class FiniteNonZero(Constant):
    """Represents non-zero rational number."""

    @property
    def raw(self) -> Fraction:
        return self._raw

    def extract_common_denominator(self) -> Tuple[int, FiniteNonZero]:
        return self.raw.denominator, FiniteNonZero(self.raw.numerator)

    def extract_common_numerator(self) -> Tuple[int, FiniteNonZero]:
        return self.raw.numerator, ONE / self.raw.denominator

    def inverse(self) -> FiniteNonZero:
        return FiniteNonZero(Fraction(self.raw.denominator,
                                      self.raw.numerator))

    def is_positive(self) -> bool:
        return self.raw > 0

    def perfect_sqrt(self) -> FiniteNonZero:
        return FiniteNonZero(Fraction(perfect_sqrt(self.raw.numerator),
                                      perfect_sqrt(self.raw.denominator)))

    def significant_digits_count(self) -> int:
        return digits_count(self._raw.limit_denominator(1).numerator)

    def square(self) -> FiniteNonZero:
        return FiniteNonZero(self.raw * self.raw)

    _raw: Fraction

    __slots__ = '_raw',

    def __init__(self, raw: RawConstant) -> None:
        assert raw and (isinstance(raw, Rational) or math.isfinite(raw)), raw
        self._raw = Fraction(raw)

    @overload
    def __add__(self, other: RawConstant) -> Union[FiniteNonZero, Infinite]:
        ...

    @overload
    def __add__(self, other: FiniteNonZero) -> FiniteNonZero:
        ...

    @overload
    def __add__(self, other: Zero) -> FiniteNonZero:
        ...

    @overload
    def __add__(self, other: Any) -> Any:
        ...

    def __add__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (to_constant(self.raw + other.raw)
                if isinstance(other, FiniteNonZero)
                else (other + self
                      if isinstance(other, Expression)
                      else NotImplemented))

    def __bool__(self) -> bool:
        return bool(self.raw)

    def __hash__(self) -> int:
        return hash(self.raw)

    @overload
    def __mul__(self, other: FiniteNonZero) -> FiniteNonZero:
        ...

    @overload
    def __mul__(self, other: RawFinite) -> Union[FiniteNonZero, Zero]:
        ...

    @overload
    def __mul__(self, other: RawUnbound) -> Union[FiniteNonZero, Infinite]:
        ...

    @overload
    def __mul__(self, other: Zero) -> Zero:
        ...

    @overload
    def __mul__(self, other: Any) -> Any:
        ...

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (FiniteNonZero(self.raw * other.raw)
                if isinstance(other, FiniteNonZero)
                else (other
                      if isinstance(other, Zero)
                      else (other * self
                            if isinstance(other, Expression)
                            else NotImplemented)))

    def __neg__(self) -> FiniteNonZero:
        return FiniteNonZero(-self.raw)

    @overload
    def __radd__(self, other: RawConstant) -> Union[FiniteNonZero, Infinite]:
        ...

    @overload
    def __radd__(self, other: Any) -> Any:
        ...

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (other + self
                if isinstance(other, Expression)
                else NotImplemented)

    __repr__ = generate_repr(__init__)

    @overload
    def __rmul__(self, other: RawConstant) -> FiniteNonZero:
        ...

    @overload
    def __rmul__(self, other: Zero) -> Zero:
        ...

    @overload
    def __rmul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (other
                if isinstance(other, Zero)
                else (other * self
                      if isinstance(other, Expression)
                      else NotImplemented))


Finite = Union[FiniteNonZero, Zero]

ZERO, ONE = Zero(), FiniteNonZero(1)


class Infinite(Constant):
    @property
    def raw(self) -> float:
        return positiveness_to_sign(self.is_positive()) * math.inf

    def extract_common_denominator(self) -> Tuple[int, Expression]:
        return 1, self

    def extract_common_numerator(self) -> Tuple[int, Expression]:
        return 1, self

    def inverse(self) -> Expression:
        return ZERO

    def is_positive(self) -> bool:
        return self._is_positive

    perfect_sqrt = identity

    def significant_digits_count(self) -> int:
        return 0

    def square(self) -> Expression:
        return Infinity

    __slots__ = '_is_positive',

    def __init__(self, is_positive: bool) -> None:
        self._is_positive = is_positive

    def __hash__(self) -> int:
        return hash(self.raw)

    @overload
    def __add__(self, other: Union[RawConstant, Expression]) -> Infinite:
        ...

    @overload
    def __add__(self, other: Any) -> Any:
        ...

    def __add__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._add_expression(other)
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __ge__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __ge__(self, other: Any) -> Any:
        ...

    def __ge__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((self.is_positive() or self == other)
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __gt__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __gt__(self, other: Any) -> Any:
        ...

    def __gt__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((self.is_positive() and self != other)
                if isinstance(other, Expression)
                else NotImplemented)

    def __le__(self,
               other: Union[RawConstant, Expression]) -> bool:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((not self.is_positive() or self == other)
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __lt__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __lt__(self, other: Any) -> Any:
        ...

    def __lt__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((not self.is_positive() and self != other)
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __mul__(self, other: Zero) -> NoReturn:
        ...

    @overload
    def __mul__(self, other: Union[RawConstant, Expression]) -> Infinite:
        ...

    @overload
    def __mul__(self, other: Any) -> Any:
        ...

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._mul_by_expression(other)
                if isinstance(other, Expression)
                else NotImplemented)

    def __neg__(self) -> Infinite:
        return Infinite(not self.is_positive())

    @overload
    def __radd__(self, other: Union[RawConstant, Expression]) -> Infinite:
        ...

    @overload
    def __radd__(self, other: Any) -> Any:
        ...

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._add_expression(other)
                if isinstance(other, Expression)
                else NotImplemented)

    __repr__ = generate_repr(__init__)

    @overload
    def __rmul__(self, other: Zero) -> NoReturn:
        ...

    @overload
    def __rmul__(self, other: Union[RawConstant, Expression]) -> Infinite:
        ...

    @overload
    def __rmul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._mul_by_expression(other)
                if isinstance(other, Expression)
                else NotImplemented)

    def _add_expression(self, other: Expression) -> Infinite:
        if (isinstance(other, Infinite)
                and self.is_positive() is not other.is_positive()):
            raise ArithmeticError('Sum of infinities with different signs '
                                  'is undefined.')
        return self

    def _mul_by_expression(self, other: Expression) -> Infinite:
        if not other:
            raise ArithmeticError('Multiplication of infinity by zero '
                                  'is undefined.')
        return (Infinity
                if self.is_positive() is other.is_positive()
                else -Infinity)


Infinity = Infinite(True)


@overload
def to_constant(_value: RawFinite) -> Union[FiniteNonZero, Zero]:
    ...


@overload
def to_constant(_value: RawUnbound) -> Union[FiniteNonZero, Infinite, Zero]:
    ...


def to_constant(_value: RawConstant) -> Union[FiniteNonZero, Infinite, Zero]:
    if not isinstance(_value, Rational) and math.isnan(_value):
        raise ValueError('NaN values are not supported.')
    return ((FiniteNonZero(_value) if _value else ZERO)
            if isinstance(_value, Rational) or math.isfinite(_value)
            else Infinite(_value > 0))



import platform

from setuptools import (find_packages,
                        setup)

project_base_url = 'https://github.com/lycantropos/symba/'
parameters = dict(packages=find_packages(exclude=('tests', 'tests.*')),
                  url=project_base_url,
                  download_url=project_base_url + 'archive/master.zip')
if platform.python_implementation() == 'CPython':
    from glob import glob

    from setuptools import Extension

    parameters.update(package_data={'symba': ['_symba.pyi']},
                      ext_modules=[Extension('symba._symba', glob('src/*.c'))],
                      zip_safe=False)
setup(**parameters)



from __future__ import annotations

import math
from numbers import Rational
from typing import (TYPE_CHECKING,
                    Any,
                    Tuple,
                    Union,
                    overload)

from reprit.base import generate_repr

from .constant import (ONE,
                       ZERO,
                       Constant,
                       FiniteNonZero,
                       Infinite,
                       Zero,
                       to_constant)
from .expression import Expression
from .hints import (RawConstant,
                    RawFinite,
                    RawUnbound)
from .utils import (ceil_half,
                    rational_sqrt_lower_bound,
                    rational_sqrt_upper_bound)

if TYPE_CHECKING:
    from .form import Form


class Term(Expression):
    """Represents square root of the expression."""

    @classmethod
    def from_components(cls,
                        scale: FiniteNonZero,
                        argument: Expression) -> Expression:
        if not (scale and argument):
            return ZERO
        denominator, argument = argument.extract_common_denominator()
        scale /= denominator
        argument *= denominator
        argument_perfect_sqrt = argument.perfect_sqrt()
        argument_perfect_part = argument_perfect_sqrt.square()
        result = (argument_perfect_sqrt * scale
                  if argument == argument_perfect_part
                  else (argument_perfect_sqrt
                        * cls(scale, argument / argument_perfect_part)))
        assert isinstance(result, Expression), result
        return result

    __slots__ = 'argument', 'scale'

    def __init__(self,
                 scale: FiniteNonZero,
                 argument: Expression) -> None:
        from .form import Form
        assert isinstance(argument, (FiniteNonZero, Form, Term)), argument
        self.argument, self.scale = argument, scale

    @property
    def degree(self) -> int:
        return self.argument.degree + 1

    def extract_common_denominator(self) -> Tuple[int, Term]:
        common_denominator, scale = self.scale.extract_common_denominator()
        return common_denominator, Term(scale, self.argument)

    def extract_common_numerator(self) -> Tuple[int, Term]:
        common_numerator, scale = self.scale.extract_common_numerator()
        return common_numerator, Term(scale, self.argument)

    def inverse(self) -> Term:
        scale = self.scale.inverse()
        argument: Expression
        denominator, argument = (
            self.argument.inverse().extract_common_denominator()
        )
        scale /= denominator
        argument *= denominator
        return Term(scale, argument)

    def is_positive(self) -> bool:
        return self.scale.is_positive()

    def lower_bound(self) -> RawConstant:
        return (rational_sqrt_lower_bound(self.square().lower_bound())
                if self.is_positive()
                else -(-self).upper_bound())

    def perfect_sqrt(self) -> Expression:
        return self.scale.perfect_sqrt()

    def significant_digits_count(self) -> int:
        return ceil_half(self.square().significant_digits_count())

    def square(self) -> Expression:
        return self.scale.square() * self.argument

    def upper_bound(self) -> RawConstant:
        return (rational_sqrt_upper_bound(self.square().upper_bound())
                if self.is_positive()
                else -(-self).lower_bound())

    @overload
    def __add__(self, other: RawFinite) -> Union[Form, Term]:
        ...

    @overload
    def __add__(self, other: RawUnbound) -> Union[Infinite, Form, Term]:
        ...

    @overload
    def __add__(self, other: FiniteNonZero) -> Form:
        ...

    @overload
    def __add__(self, other: Infinite) -> Infinite:
        ...

    @overload
    def __add__(self, other: Zero) -> Term:
        ...

    @overload
    def __add__(self, other: Any) -> Any:
        ...

    def __add__(self, other: Any) -> Any:
        from .form import Form
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._add_constant(other)
                if isinstance(other, Constant)
                else (Form.from_components([self, other])
                      if isinstance(other, Term)
                      else NotImplemented))

    @overload
    def __eq__(self, other: Expression) -> bool:
        ...

    @overload
    def __eq__(self, other: Any) -> Any:
        ...

    def __eq__(self, other: Any) -> Any:
        return (isinstance(other, Term)
                and self.is_positive() is other.is_positive()
                and self.square() == other.square()
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __ge__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __ge__(self, other: Any) -> Any:
        ...

    def __ge__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other <= self
                 if isinstance(other, Infinite)
                 else (not other.is_positive()
                       or other.square() <= self.square()
                       if self.is_positive()
                       else (not other.is_positive()
                             and other.square() >= self.square())))
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __gt__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __gt__(self, other: Any) -> Any:
        ...

    def __gt__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other < self
                 if isinstance(other, Infinite)
                 else ((not other.is_positive()
                        or other.square() < self.square())
                       if self.is_positive()
                       else (not other.is_positive()
                             and other.square() > self.square())))
                if isinstance(other, Expression)
                else NotImplemented)

    def __hash__(self) -> int:
        return hash((self.is_positive(), self.square()))

    @overload
    def __le__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __le__(self, other: Any) -> Any:
        ...

    def __le__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other >= self
                 if isinstance(other, Infinite)
                 else ((other.is_positive()
                        and other.square() >= self.square())
                       if self.is_positive()
                       else (other.is_positive()
                             or other.square() <= self.square())))
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __lt__(self, other: Union[RawConstant, Expression]) -> bool:
        ...

    @overload
    def __lt__(self, other: Any) -> Any:
        ...

    def __lt__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return ((other > self
                 if isinstance(other, Infinite)
                 else ((other.is_positive() and other.square() > self.square())
                       if self.is_positive()
                       else (other.is_positive()
                             or other.square() < self.square())))
                if isinstance(other, Expression)
                else NotImplemented)

    @overload
    def __mul__(self, other: FiniteNonZero) -> Term:
        ...

    @overload
    def __mul__(self, other: Infinite) -> Infinite:
        ...

    @overload
    def __mul__(self, other: RawFinite) -> Union[Term, Zero]:
        ...

    @overload
    def __mul__(self, other: RawUnbound) -> Union[Infinite, Term, Zero]:
        ...

    @overload
    def __mul__(self, other: Term) -> Union[FiniteNonZero, Form, Term]:
        ...

    @overload
    def __mul__(self, other: Zero) -> Zero:
        ...

    @overload
    def __mul__(self, other: Any) -> Any:
        ...

    def __mul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._multiply_by_constant(other)
                if isinstance(other, Constant)
                else (self._multiply_by_term(other)
                      if isinstance(other, Term)
                      else NotImplemented))

    def __neg__(self) -> Term:
        return Term(-self.scale, self.argument)

    @overload
    def __radd__(self, other: Union[RawConstant, Expression]) -> Expression:
        ...

    @overload
    def __radd__(self, other: Any) -> Any:
        ...

    def __radd__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._add_constant(other)
                if isinstance(other, Constant)
                else NotImplemented)

    __repr__ = generate_repr(__init__)

    @overload
    def __rmul__(self, other: FiniteNonZero) -> Term:
        ...

    @overload
    def __rmul__(self, other: RawFinite) -> Union[Term, Zero]:
        ...

    @overload
    def __rmul__(self, other: RawUnbound) -> Union[Infinite, Term, Zero]:
        ...

    @overload
    def __rmul__(self, other: Zero) -> Zero:
        ...

    @overload
    def __rmul__(self, other: Any) -> Any:
        ...

    def __rmul__(self, other: Any) -> Any:
        if isinstance(other, (Rational, float)):
            other = to_constant(other)
        return (self._multiply_by_constant(other)
                if isinstance(other, Constant)
                else NotImplemented)

    def __str__(self) -> str:
        return ((''
                 if self.scale == ONE
                 else ('-'
                       if self.scale == -ONE
                       else '{} * '.format(self.scale)))
                + 'sqrt({})'.format(self.argument))

    def _add_constant(self, other: Constant) -> Expression:
        from .form import Form
        return (Form([self], other)
                if isinstance(other, FiniteNonZero)
                else (other
                      if isinstance(other, Infinite)
                      else self))

    def _multiply_by_constant(self, other: Constant) -> Expression:
        return (Term(self.scale * other, self.argument)
                if isinstance(other, FiniteNonZero)
                else ((other if self.is_positive() else -other)
                      if isinstance(other, Infinite)
                      else other))

    def _multiply_by_term(self, other: Term) -> Expression:
        scale = self.scale * other.scale
        if self.argument == other.argument:
            return scale * self.argument
        argument, other_argument = ((self.argument, other.argument)
                                    if self.argument < other.argument
                                    else (other.argument, self.argument))
        if not other_argument % argument:
            return argument * Term.from_components(scale,
                                                   other_argument / argument)
        elif (isinstance(argument, FiniteNonZero)
              and isinstance(other_argument, FiniteNonZero)):
            arguments_gcd = math.gcd(argument.raw.numerator,
                                     other_argument.raw.numerator)
            argument /= arguments_gcd
            other_argument /= arguments_gcd
            arguments_gcd_expression = to_constant(arguments_gcd)
            assert isinstance(arguments_gcd_expression, FiniteNonZero), (
                arguments_gcd_expression
            )
            scale *= arguments_gcd_expression
        return Term.from_components(scale, argument * other_argument)



from typing import Union as _Union

from .core import expression as _expression
from .core.constant import (ONE as _ONE,
                            Infinite as _Infinite,
                            to_constant as _to_constant)
from .core.hints import RawConstant as _RawConstant
from .core.term import Term as _Term

Expression = _expression.Expression


def sqrt(argument: _Union[_RawConstant, Expression]) -> Expression:
    """
    Returns square root of the argument:
        exact if it is a perfect square, symbolic instead.

    >>> sqrt(0) == 0
    True
    >>> sqrt(1) == 1
    True
    >>> square_root_of_two = sqrt(2)
    >>> square_root_of_two ** 2 == 2
    True
    >>> 1 < square_root_of_two < 2
    True
    """
    if argument < 0:
        raise ValueError('Argument should be non-negative.')
    expression = (argument
                  if isinstance(argument, Expression)
                  else _to_constant(argument))
    return (expression
            if isinstance(expression, _Infinite)
            else _Term.from_components(_ONE, expression))
